/* 
   filebrowser_next.c - NeXTifyed file browser
   Copyright (C) 1999 Sasha Vasko

   Modifyed version of :
   filebrowser.c - draws an interactive file browser
   Copyright (C) 1996, 1997 Paul Sheer

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
 */

#include <config.h>
#include <stdio.h>
#include <my_string.h>
#include <stdlib.h>
#include <stdarg.h>

#include <sys/types.h>
#include <sys/stat.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xresource.h>
#include "lkeysym.h"

#include "stringtools.h"
#include "app_glob.c"

#include "coolwidget.h"
#include "editcmddef.h"

#include "mad.h"

/*****************************************************************/
/*     Miscellaneous read only text widget for dynamic text      */
/*****************************************************************/

void render_dyn_text (CWidget * wdt)
{
    Window win = wdt->winid;
    char text[1024], *p, *q;
    int y, x, w = wdt->width;
    int color = 0;
    int new_line = 1;

    CSetColor (COLOR_LIGHT);
    CRectangle (win, 1, 1, w - 2, wdt->height - 2);
    CSetColor (COLOR_BLACK);

    y = 1;			/* bevel */
    if ((q = wdt->text) == NULL)
	return;
    CPushFont ("widget", 0);
    for (;;) {
	long len;

	CSetColor (color_palette (color));
	/* looking for control characters */
	for (p = q; *p && *p >= ' '; p++);

	len = min ((unsigned long) p - (unsigned long) q, 1023);

	if (new_line) {
	    if (wdt->options & TEXT_CENTRED)
		x = (w - (TEXT_RELIEF + 1) * 2 - CTextWidth (q, len)) / 2;
	    else
		x = 0;
	    new_line = 0;
	}
	if (!*p) {		/* last line */
	    drawstring_xy (win, TEXT_RELIEF + 1 + x, TEXT_RELIEF + y, q);
	    break;
	} else {
	    memcpy (text, q, len);
	    text[len] = 0;
	    drawstring_xy (win, TEXT_RELIEF + 1 + x,
			   TEXT_RELIEF + y, text);
	    x += CTextWidth (q, len);
	}

	if (*p == '\n') {
	    new_line = 1;
	    y += FONT_PIX_PER_LINE;	/* next line */
	} else if (*p == 28)
	    color = (int) '\n';
	else
	    color = *p;

	color %= 27;
	q = p + 1;
    }
    CPopFont ();
    render_bevel (win, 0, 0, w - 1, wdt->height - 1, 1, 1);
}


/*-----------------------------------------------------------------------*/
int eh_dyn_text (CWidget * w, XEvent * xevent, CEvent * cwevent)
{
    switch (xevent->type) {
    case Expose:
	if (!xevent->xexpose.count)
	    render_dyn_text (w);
	break;
    }
    return 0;
}

CWidget *CDrawDynText (const char *identifier, Window parent,
		  int x, int y, int width, int rows, const char *fmt,...)
{
    va_list pa;
    char *str;
    int w, h;
    CWidget *wdt;

    va_start (pa, fmt);
    str = vsprintf_alloc (fmt, pa);
    va_end (pa);

    CPushFont ("widget", 0);
    CTextSize (&w, &h, str);
    if (width != AUTO_WIDTH && width)
	w = width;
    else
	w += TEXT_RELIEF * 2 + 2;

    if (rows != AUTO_HEIGHT && rows) {
	CTextSize (0, &h, "Ty~g$V|&^.,/_'I}[@");
	h = (h + FONT_OVERHEAD) * rows;
    }
    h += TEXT_RELIEF * 2 + 2;

    wdt = CSetupWidget (identifier, parent, x, y,
			w, h, C_TEXT_WIDGET, INPUT_EXPOSE, COLOR_FLAT, 0);
    wdt->text = (char *) strdup (str);
    wdt->eh = eh_dyn_text;
    free (str);
    set_hint_pos (x + w + WIDGET_SPACING, y + h + WIDGET_SPACING);
    CPopFont ();
    return wdt;
}

CWidget *CRedrawDynText (const char *identifier, const char *fmt,...)
{
    va_list pa;
    char *str;
    CWidget *wdt;
    int w, h;

    wdt = CIdent (identifier);
    if (!wdt)
	return 0;

    va_start (pa, fmt);
    str = vsprintf_alloc (fmt, pa);
    va_end (pa);

    if (wdt->text)
	free (wdt->text);
    wdt->text = (char *) strdup (str);

    render_dyn_text (wdt);
    free (str);
    return wdt;
}


/******************************************************************/
/*                   NeXT Directory Tree stuff                    */
/******************************************************************/
typedef struct NeXTDir {
    struct NeXTDir *next, *prev;

    char *name;
    char *filter;
    long cursor;
    long current;
    long firstline;
    struct file_entry *list;

} NeXTDir;

typedef struct NeXTDirTree {
    NeXTDir *first, *last, *selected;
    NeXTDir *discarded;		/* when we chdir - it will store previos tail */
    int pos;
    int numdirs;
    int numpanes;
    char *path;
    int path_length;

    int dirty;			/* can't change dirs - need to redraw widgets first */
} NeXTDirTree;

extern Bool is_directory (struct file_entry *directentry, int line_number);

NeXTDir *create_dir_elem (char *name, int name_len)
{
    NeXTDir *dir;
    dir = (NeXTDir *) calloc (sizeof (NeXTDir), 1);
    dir->prev = dir->next = NULL;
    dir->filter = (char *) strdup ("*");
    dir->current = -1;
    dir->list = NULL;
    if (name == NULL || name_len == 0) {	/* root directory then */
	dir->name = (char *) strdup ("");
    } else {
	dir->name = malloc (name_len + 1);
	strncpy (dir->name, name, name_len);
	dir->name[name_len] = '\0';
    }
    return dir;
}

void destroy_dir_elem (NeXTDir ** dir)
{
    if (dir) {
	if (*dir) {
	    if ((*dir)->filter)
		free ((*dir)->filter);
	    if ((*dir)->name)
		free ((*dir)->name);
	    if ((*dir)->list)
		free ((*dir)->list);
	    memset (*dir, 0x00, sizeof (NeXTDir));	/* just in case */
	    free (*dir);
	    *dir = NULL;
	}
    }
}

NeXTDirTree *create_dir_tree (const char *path, int numpanes)
{
    NeXTDirTree *tree = NULL;
    char *ptr, *next;
    int i;

    tree = (NeXTDirTree *) calloc (sizeof (NeXTDirTree), 1);

    if (ptr == NULL) {		/* defaulting to the root */
	tree->path_length = 1;
	tree->path = (char *) strdup ("/");
	tree->first = tree->last = tree->selected = create_dir_elem ("", 0);
    } else {
	tree->path_length = strlen (path);
	tree->path = malloc (tree->path_length + 1);
	strcpy (tree->path, path);

	ptr = tree->path;

	if (*ptr == '/') {
	    tree->first = tree->last = tree->selected = create_dir_elem ("", 0);
	    tree->numdirs++;
	}
	while (*ptr) {
	    int len = 0;
	    while (*(ptr + len) && *(ptr + len) != '/')
		len++;
	    if (len) {
		tree->selected = create_dir_elem (ptr, len);
		if (tree->last == NULL)
		    tree->first = tree->last = tree->selected;
		else {
		    tree->selected->prev = tree->last;
		    tree->last->next = tree->selected;
		    tree->last = tree->selected;
		}
		tree->numdirs++;
	    }
	    ptr += len;
	    if (*ptr)
		ptr++;		/* skipping / */
	}
    }
    tree->numpanes = numpanes;
    tree->discarded = NULL;

    if (tree->selected == NULL)
	tree->selected = tree->last;
    tree->pos = tree->numdirs - tree->numpanes;
    if (tree->pos < 0)
	tree->pos = 0;
    return tree;
}

void destroy_dir_chain (NeXTDir ** start)
{
    NeXTDir *dir;
    for (dir = *start; dir; dir = *start) {
	*start = dir->next;
	destroy_dir_elem (&dir);
    }
}

void destroy_dir_tree (NeXTDirTree ** tree)
{
    if (tree) {
	destroy_dir_chain (&((*tree)->first));
	destroy_dir_chain (&((*tree)->discarded));
	free (*tree);
	*tree = NULL;
    }
}

char *get_dir_path (NeXTDirTree * tree, NeXTDir * dir, char *buffer)
{
    register NeXTDir *d;
    register char *ptr1 = buffer, *ptr2;
    if (!ptr1 || !tree || !dir)
	return NULL;

    for (d = tree->first; d && d != dir; d = d->next) {
	*(ptr1++) = '/';
	ptr2 = d->name;
	if (*ptr2)
	    while (*ptr2)
		*(ptr1++) = *(ptr2++);
	else			/* we don't really want //, even thou it is not a violation */
	    ptr1--;
    }
    *ptr1 = '\0';
    return buffer;
}

struct file_entry *read_dir_filelist (NeXTDir * dir, char *path)
{
    if (dir && path) {
	register int k;
	char *next_dir = NULL;
	CHourGlass (CFirstWindow);
	if (dir->list)
	    free (dir->list);
	dir->list = get_file_entry_list (path, FILELIST_ALL_FILES, (dir->filter) ? dir->filter : "*");
	if (dir->next)
	    next_dir = dir->next->name;
	dir->current = -1;
	if (dir->list && next_dir) {
	    for (k = 0; !(dir->list[k].options & FILELIST_LAST_ENTRY); k++)
		if (is_directory (dir->list, k))
		    if (strcmp (dir->list[k].name, next_dir) == 0) {
			dir->current = k;
			break;
		    }
	}
	dir->cursor = dir->firstline = (dir->current >= 0) ? dir->current : 0;
	CUnHourGlass (CFirstWindow);
    }
    return dir->list;
}
void update_tree_integrity (NeXTDirTree * tree, NeXTDir * selected)
{
    NeXTDir *t;
    tree->selected = NULL;
    tree->numdirs = 0;
    tree->path_length = 0;

    for (t = tree->first; t; t = t->next) {
	tree->last = t;
	if (selected == t)
	    tree->selected = selected;
	tree->numdirs++;
	tree->path_length += 1 + strlen (t->name);
    }

    tree->path_length++;

    tree->pos = tree->numdirs - tree->numpanes;
    if (tree->pos < 0)
	tree->pos = 0;

    if (tree->selected == NULL)
	tree->selected = tree->last;
}

void change_dir (NeXTDirTree * tree, NeXTDir * dir, long item_num)
{
    struct file_entry *item;
    NeXTDir *tmp;

    if (!tree || tree->dirty || !dir || item_num < 0)
	return;
    item = &(dir->list[item_num]);

    if (item->name[0] == '.') {
	if (item->name[1] == '\0')
	    return;
	else if (item->name[1] == '.' && item->name[2] == '\0') {	/* poping Up */
	    if (dir->prev == NULL)
		return;
	    dir = dir->prev;
	    if (tree->discarded)
		destroy_dir_chain (&(tree->discarded));
	    tree->discarded = dir->next;
	    dir->next = NULL;
	    tree->dirty = 1;
	    update_tree_integrity (tree, dir);
	    return;
	}
    }
    if (tree->discarded) {
	if (tree->discarded->prev == dir &&
	    strcmp (tree->discarded->name, item->name) == 0) {
	    tmp = dir->next;
	    dir->next = tree->discarded;
	    tree->discarded = tmp;
	} else {
	    destroy_dir_chain (&(tree->discarded));
	    tree->discarded = dir->next;
	    dir->next = NULL;
	}
    } else {
	tree->discarded = dir->next;
	dir->next = NULL;
    }

    if (dir->next == NULL) {
	dir->next = create_dir_elem (item->name, strlen (item->name));
	dir->next->prev = dir;
    }
    tree->dirty = 1;

    dir->current = item_num;
    /* taking some precautions so not to destroy tree integrity */
    update_tree_integrity (tree, dir);
}

char *get_full_filename (NeXTDirTree * tree, NeXTDir * dir, long item)
{
    char *buffer, *path_tail;
    if (!tree || item < 0 || !dir || !(dir->list))
	return NULL;
    buffer = malloc (tree->path_length + 1 + strlen (dir->list[item].name) + 1);
    get_dir_path (tree, dir, buffer);
    path_tail = buffer + strlen (buffer);
    sprintf (path_tail, "/%s/%s", dir->name, dir->list[item].name);
    return buffer;
}

char *commit (NeXTDirTree * tree, NeXTDir * dir)
{				/* when user presses Enter on selected item or dobleclicks */
    if (tree) {
	if (dir) {
	    if (dir->list && dir->cursor >= 0) {
		if (is_directory (dir->list, dir->cursor))
		    change_dir (tree, dir, dir->cursor);
		else
		    return get_full_filename (tree, dir, dir->cursor);
	    }
	}
    }
    return NULL;
}

/******************************************************************/
/*                   NeXT Browser widget stuff                    */
/******************************************************************/

#define FILEBROWSER_PANES 		2
#define STANDALONE_FILEBROWSER_PANES 	3

static char *mime_majors[3] =
{"url", "text", 0};

typedef struct NeXTFileBrowser {
    int numpanes;
    int focused;

#define MAX_PARTIAL_NAME 32
    char partial_name[MAX_PARTIAL_NAME];
    int partial_end;

    char **labels;
    char **lists;
    char *hscroll;
    char *fileattr;
    char *name;
    char *filter;
    NeXTDirTree *tree;
} NeXTFileBrowser;

void clear_partial_buffer (NeXTFileBrowser * browser)
{
    if (browser) {
	browser->partial_end = 0;
    }
}


extern void CSetFilelistPosition (const char *identifier, long current, long cursor, long firstline);

void dir_save_position (NeXTDir * dir, long cursor, long current, long firstline)
{
    dir->cursor = cursor;
    dir->current = current;
    dir->firstline = firstline;
}

void destroy_next_filebrowser (CWidget * w)
{
    if (w->hook) {
	NeXTFileBrowser *browser = (NeXTFileBrowser *) w->hook;
	int i;
	destroy_dir_tree (&(browser->tree));
	if (browser->labels) {
	    for (i = 0; i < browser->numpanes; i++)
		if (browser->labels[i])
		    free (browser->labels[i]);
	    free (browser->labels);
	}
	if (browser->lists) {
	    for (i = 0; i < browser->numpanes; i++)
		if (browser->lists[i])
		    free (browser->lists[i]);
	    free (browser->lists);
	}
	if (browser->fileattr)
	    free (browser->fileattr);
	if (browser->name)
	    free (browser->name);
	if (browser->filter)
	    free (browser->filter);
	/* just in case */
	memset (browser, 0x00, sizeof (NeXTFileBrowser));

	free (browser);
	w->hook = 0;
    }
}

NeXTDir *get_dir_by_num (NeXTDirTree * tree, int num)
{
    NeXTDir *dir = NULL;
    register int i;
    if (tree)
	dir = tree->first;
    for (i = 0; dir && i < num; i++)
	dir = dir->next;
    return dir;
}

void update_filter (NeXTFileBrowser * browser, NeXTDir * dir);

void link_browser_to_data (NeXTFileBrowser * browser, int start)
{
    NeXTDir *dir;
    char *path = NULL, *path_tail;
    int i, todo = browser->numpanes;
    if (browser == NULL)
	return;
    if (start < 0 || start >= browser->numpanes)
	start = 0;

    if (start == 0)
	browser->tree->dirty = 0;

    if (browser->tree) {
	todo -= start;
	start += browser->tree->pos;
	dir = get_dir_by_num (browser->tree, start);
    } else
	dir = NULL;

    if (dir) {
	path = malloc (browser->tree->path_length + 1);
	get_dir_path (browser->tree, dir, path);
	path_tail = path + strlen (path);
    }
    for (i = 0; i < todo; i++) {
	if (!dir) {
	    CRedrawDynText (browser->labels[i], "");
	    CRedrawFilelist (browser->lists[i], 0, 0);
	} else {

	    if (dir->name) {
		*(path_tail++) = '/';
		strcpy (path_tail, dir->name);
		path_tail += strlen (dir->name);
		CRedrawDynText (browser->labels[i], "%s/%s", dir->name, dir->filter);
	    } else
		CRedrawDynText (browser->labels[i], "/%s", dir->filter);	/* root */

	    if (dir->list == NULL)
		read_dir_filelist (dir, path);
	    CRedrawFilelist (browser->lists[i], dir->list, 0);
	    CSetFilelistPosition (browser->lists[i], dir->current, dir->cursor, dir->firstline);
	    dir = dir->next;
	}
    }
}

void update_scrollbar (CWidget * scrollbar, NeXTFileBrowser * browser, int redraw)
{
    scrollbar->firstline = (double) 65535.0 *browser->tree->pos / (browser->tree->numdirs ? browser->tree->numdirs : 1);
    scrollbar->numlines = (double) 65535.0 *browser->numpanes / (browser->tree->numdirs ? browser->tree->numdirs : 1);
    if (redraw)
	render_scrollbar (scrollbar);
}

void link_scrollbar_to_NeXT_browser (CWidget * scrollbar, CWidget * w, XEvent * xevent, CEvent * cwevent, int whichscrbutton)
{
    /* fix me : add stuf */
    int redraw = 0, count, c;
    int new_first = w->firstline;
    static int r = 0;
    NeXTFileBrowser *browser;
    NeXTDir *dir;

    if (w->hook == NULL)
	return;
    browser = (NeXTFileBrowser *) (w->hook);
    new_first = browser->tree->pos;

    if ((xevent->type == ButtonRelease || xevent->type == MotionNotify) && whichscrbutton == 3) {
	new_first = (double) scrollbar->firstline * browser->tree->numdirs / 65535.0;
    } else if (xevent->type == ButtonPress && (cwevent->button == Button1 || cwevent->button == Button2)) {
	switch (whichscrbutton) {
	case 1:
	    new_first -= browser->numpanes - 1;
	    break;
	case 2:
	    new_first--;
	    break;
	case 5:
	    new_first++;
	    break;
	case 4:
	    new_first += browser->numpanes - 1;
	    break;
	}
    }
    if (new_first > browser->tree->numdirs - browser->numpanes)
	new_first = browser->tree->numdirs - browser->numpanes;
    if (new_first < 0)
	new_first = 0;

    redraw = (new_first != browser->tree->pos);
    browser->tree->pos = new_first;

    if (xevent->type == ButtonRelease ||
	(redraw && !CCheckWindowEvent (xevent->xany.window, ButtonReleaseMask | ButtonMotionMask, 0)))
	link_browser_to_data (browser, 0);

    update_scrollbar (w->hori_scrollbar, browser, 0);
}

void update_input (NeXTFileBrowser * browser, NeXTDir * dir)
{
    char *text = NULL;
    CWidget *inp_w = CIdent (browser->name);

    clear_partial_buffer (browser);

    text = get_full_filename (browser->tree, dir, dir->cursor);

    CDrawTextInput (browser->name, inp_w->parentid, inp_w->x, inp_w->y,
		    inp_w->width, inp_w->height, 256, text);
    if (text)
	free (text);

}

void filter_changed (NeXTFileBrowser * browser, NeXTDir * dir)
{
    CWidget *inp_w = CIdent (browser->filter);
    char *new_filter = "*";
    if (inp_w->text)
	new_filter = inp_w->text;

    if (strcmp (dir->filter, new_filter)) {
	struct file_entry *list;
	if (dir->filter)
	    free (dir->filter);
	dir->filter = (char *) strdup (new_filter);
	list = dir->list;
	dir->list = NULL;
	link_browser_to_data (browser, browser->focused - browser->tree->pos);
	/* now after all filelists has updated its information 
	   we no longer need old data */
	if (list)
	    free (list);
    }
}

void update_filter (NeXTFileBrowser * browser, NeXTDir * dir)
{
    CWidget *inp_w = CIdent (browser->filter);

    if (strcmp (dir->filter, inp_w->text) == 0)
	return;
    if (strlen (inp_w->text))
	CAddToTextInputHistory (browser->name, inp_w->text);

    CDrawTextInput (browser->filter, inp_w->parentid, inp_w->x, inp_w->y,
		    inp_w->width, inp_w->height, 256, dir->filter);

}

/* returns 0 on fail */
static int goto_partial_file_name (NeXTFileBrowser * browser, NeXTDir * dir, int new_char)
{
    int success = 0;
    if (browser && dir && dir->list && browser->partial_end < MAX_PARTIAL_NAME - 1) {
	register int i;
	if (new_char != 0) {
	    browser->partial_name[browser->partial_end] = (char) new_char;
	    browser->partial_end++;
	} else if (browser->partial_end <= 0)
	    return success;
	else
	    browser->partial_end--;

	for (i = 0; !(dir->list[i].options & FILELIST_LAST_ENTRY); i++)
	    if (strncmp (dir->list[i].name, browser->partial_name, browser->partial_end) == 0) {
		dir->cursor = i;
		dir->firstline = dir->cursor;
		CSetFilelistPosition (browser->lists[browser->focused], dir->current, dir->cursor, dir->firstline);
		success++;
		break;
	    }
	if (!success)
	    browser->partial_end--;

    }
    return success;
}

Window draw_file_browser (const char *identifier, Window parent, int x, int y,
	      const char *directory, const char *file, const char *label)
{
    CWidget *w, *tmp;
    int y2, x2, x3, btn_width, btn_height;
    int text_w1, text_w2;
    Window win;
    NeXTFileBrowser *browser;
    int i;
    char *ident, *ctrl_name;
    int panes = FILEBROWSER_PANES;

    if (strcmp (identifier, "browser") == 0)
	panes = STANDALONE_FILEBROWSER_PANES;

    if (parent == CRoot)
	win = CDrawMainWindow (identifier, label);
    else
	win = CDrawHeadedDialog (identifier, parent, x, y, label);

    w = CIdent (identifier);
    w->options |= WINDOW_ALWAYS_RAISED;
    CHourGlass (CFirstWindow);
/* allocating, initializing and storing our data */
    browser = (NeXTFileBrowser *) calloc (sizeof (NeXTFileBrowser), 1);
    w->hook = browser;
    w->destroy = destroy_next_filebrowser;
    browser->tree = create_dir_tree (directory, panes);
    browser->numpanes = 0;
    CUnHourGlass (CFirstWindow);

    if (browser->tree->numdirs == 0) {
	CErrorDialog (parent, 20, 20, _ (" File browser "), _ (" Unable to read directory "));
	CDestroyWidget (identifier);
	return None;
    }
    browser->labels = (char **) calloc (sizeof (char *), panes);
    browser->lists = (char **) calloc (sizeof (char *), panes);
    browser->numpanes = panes;

    CGetHintPos (&x, &y);
    x += NEXT_SPACING;
    y += NEXT_SPACING;
#ifdef NEXT_LOOK
#define FILE_BOX_WIDTH (FONT_MEAN_WIDTH * 32 + 7)
#else
#define FILE_BOX_WIDTH (FONT_MEAN_WIDTH * 24 + 7)
#endif
    /* we'll use it to construct our widget names */
    ident = malloc (strlen (identifier) + 20);
    strcpy (ident, identifier);
    ctrl_name = ident + strlen (identifier);
    y2 = 0;
    for (i = 0; i < browser->numpanes; i++) {
	CGetHintPos (&x2, 0);
	x2 += NEXT_SPACING;
	sprintf (ctrl_name, ".label%3.3d", i);
	browser->labels[i] = (char *) strdup (ident);
	tmp = CDrawDynText (ident, win, x2, y, FILE_BOX_WIDTH, 1, "");
	tmp->options |= TEXT_CENTRED;

	/* only first time */
	if (y2 == 0) {
	    CGetHintPos (0, &y2);
	    /* y2+=NEXT_SPACING ; */
	}
	sprintf (ctrl_name, ".flist%3.3d", i);
	browser->lists[i] = (char *) strdup (ident);
	tmp = CDrawFilelist (ident, win, x2, y2,
			     FILE_BOX_WIDTH, FONT_PIX_PER_LINE * 10, 0, 0, NULL, TEXTBOX_FILE_LIST);
	tmp->position |= POSITION_HEIGHT;
	xdnd_set_type_list (CDndClass, tmp->winid, xdnd_typelist_send[DndFiles]);
	/* Toolhint */
	CSetToolHint (ident, _ ("Double click to enter dir or open file."));
    }
    /* we want to automagically resize rightmost filelist */
    tmp->position |= POSITION_WIDTH;
    if (tmp->vert_scrollbar)
	tmp->vert_scrollbar->position |= POSITION_RIGHT;
    tmp = CIdent (browser->labels[browser->numpanes - 1]);
    tmp->position |= POSITION_WIDTH;

    /* the right end of dialog */
    CGetHintPos (&x3, &y2);

    strcpy (ctrl_name, ".hsc");
    i = (double) 65535.0 *browser->tree->numpanes / (browser->tree->numdirs ? browser->tree->numdirs : 1);
    w->hori_scrollbar = CDrawHorizontalScrollbar (ident, win,
			x, y2 + NEXT_SPACING, x3 - x, AUTO_HEIGHT, i, i);
    CSetScrollbarCallback (ident, w->ident, link_scrollbar_to_NeXT_browser);
    w->hori_scrollbar->position = POSITION_BOTTOM | POSITION_WIDTH;
    update_scrollbar (w->hori_scrollbar, browser, 1);

    CPushFont ("widget", 0);

    CGetHintPos (0, &y2);
    strcpy (ctrl_name, ".fileattr");
    browser->fileattr = (char *) strdup (ident);
    tmp = CDrawDynText (ident, win, x, y2 + NEXT_SPACING, x3 - x, 3, "");
    tmp->position |= POSITION_FILL | POSITION_BOTTOM;

    CTextSize (&text_w1, 0, "Name : ");
    CTextSize (&text_w2, 0, "Filter : ");
/* filename input stuff: */
    CGetHintPos (0, &y2);
    y2 += NEXT_SPACING;
    /* label */
    strcpy (ctrl_name, ".namex");
    x2 = x;
    if (text_w1 < text_w2)
	x2 += text_w2 - text_w1;
    tmp = CDrawText (ident, win, x2, y2, _ ("Name : "));
    tmp->position |= POSITION_BOTTOM;
    /* we want filter and name inputs to be right underneath of each other */
    x2 += text_w1 + NEXT_SPACING;
    /* input */
    strcpy (ctrl_name, ".finp");
    browser->name = (char *) strdup (ident);
    tmp = CDrawTextInput (ident, win, x2, y2, AUTO_WIDTH, AUTO_HEIGHT, 256, file);
    tmp->position |= POSITION_FILL | POSITION_BOTTOM;
    /* Toolhint */
    CSetToolHint (ident, _ ("Filename of the file to load."));
    /* DnD stuff */
    xdnd_set_type_list (CDndClass, tmp->winid, xdnd_typelist_send[DndFile]);
    tmp->funcs->types = DndFile;
    tmp->funcs->mime_majors = mime_majors;

/* file filter input stuff */
    CGetHintPos (0, &y2);
    y2 += NEXT_SPACING;
    /* label */
    x2 = x;
    if (text_w2 < text_w1)
	x2 += text_w1 - text_w2;
    strcpy (ctrl_name, ".filtx");
    (CDrawText (ident, win, x2, y2, _ ("Filter :")))->position |= POSITION_BOTTOM;
    x2 += text_w2 + NEXT_SPACING;
    /* input (we already know where to place it - right under filename input ) */
    strcpy (ctrl_name, ".filtinp");
    browser->filter = (char *) strdup (ident);
    tmp = CDrawTextInput (ident, win, x2, y2, AUTO_WIDTH, AUTO_HEIGHT, 256, TEXTINPUT_LAST_INPUT);
    tmp->position |= POSITION_FILL | POSITION_BOTTOM;
    /* Toolhint */
    CSetToolHint (ident, _ ("List only files matching this shell filter"));

/* buttons stuff */
    CGetHintPos (0, &y2);
    y2 += NEXT_SPACING * 2;
    /* determining buttons size */
    CTextSize (&btn_width, &btn_height, " Cancel ");
    btn_height += 5 + BUTTON_RELIEF * 2;
    btn_width += 4 + BUTTON_RELIEF * 2;
    /* cancel button */
    x2 = x3 - (btn_width + NEXT_SPACING) * 2 - NEXT_SPACING;
    strcpy (ctrl_name, ".cancel");
    tmp = CDrawButton (ident, win, x2, y2, btn_width, btn_height, " Cancel ");
    tmp->position |= POSITION_RIGHT | POSITION_BOTTOM;
    CSetToolHint (ident, _ ("Abort this dialog, Escape"));
    /* ok button */
    x2 += btn_width + NEXT_SPACING * 2;
    strcpy (ctrl_name, ".ok");
    tmp = CDrawButton (ident, win, x2, y2, btn_width, btn_height, "   Ok   ");
    tmp->position |= POSITION_RIGHT | POSITION_BOTTOM;
    CSetToolHint (ident, _ ("Accept, Enter"));

    CPopFont ();

/* all done - no longer need that : */
    free (ident);

/* make us real ! */
    link_browser_to_data (browser, 0);
    /* that will be our width/height : */
    y2 += btn_height + NEXT_SPACING * 2;
    reset_hint_pos (x3, y2);
    CSetSizeHintPos (identifier);
    /* map us */
    CMapDialog (identifier);
    /* enable resizing */
    CSetWindowResizable (identifier, FONT_MEAN_WIDTH * 40, min (FONT_PIX_PER_LINE * 5 + 230, w->height), 1600, 1200);	/* minimum and maximum sizes */

    return win;
}

/* options */
#define GETFILE_GET_DIRECTORY		1
#define GETFILE_GET_EXISTING_FILE	2
#define GETFILE_BROWSER			4

void input_insert (CWidget * w, int c);

static char *file_error (void)
{
    if (errno) {
	char *error_msg;
#ifdef HAVE_STRERROR
	error_msg = _ (strerror (errno));
#else
	extern int sys_nerr;
	extern char *sys_errlist[];
	if ((0 <= errno) && (errno < sys_nerr))
	    error_msg = _ (sys_errlist[errno]);
	else
/* The returned value, 'errno' has an unknown meaning */
	    error_msg = _ ("strange errno");
#endif
	return catstrs ("[", error_msg, "]", 0);
    }
    return "";
}

void run_cmd (const char *fmt,...)
{
    signal (SIGCHLD, SIG_IGN);
    if (!fork ()) {		/* child process */
	va_list pa;
	char *str;

	va_start (pa, fmt);
	str = vsprintf_alloc (fmt, pa);
	va_end (pa);

	execlp ("/bin/sh", "/bin/sh", "-c", str, (char *) NULL);

	/* if all is fine then the thread will exit here */
	/* so displaying error if not                    */
	fprintf (stderr, "\nBad luck running command [%s].\n", str);
	exit (0);		/*thread completed */
    }
}

char *empty_line = "";

char *item_selected (const char *identifier, NeXTFileBrowser * browser, NeXTDir * dir)
{
    char *ptr = NULL;
    CWidget *w;

    if (!browser || !identifier)
	return 0;		/* something terrible has happen ! */
    if (dir == NULL) {
	if ((w = CIdent (browser->name)) && w->text)
	{
	    strdup (ptr = w->text);
	    if (strlen (w->text))
		CAddToTextInputHistory (browser->name, w->text);
	}
    } else {
	ptr = commit (browser->tree, dir);
	if (browser->tree->dirty)
	    link_browser_to_data (browser, 0);
    }

    if (ptr == NULL)
	ptr = empty_line;
    else if (strcmp (identifier, "browser") == 0) {
	struct file_entry *item = &(dir->list[dir->cursor]);

	if ((item->stat.st_mode & (S_IXUSR | S_IXGRP | S_IXOTH))) {
	    run_cmd (ptr);
	} else {
	    static char *default_editor = "smalledit";
	    static char *editor = NULL;
	    if (editor == NULL) {
		editor = getenv ("EDITOR");
		if (editor == NULL)
		    editor = default_editor;
	    }
	    run_cmd ("%s %s", editor, ptr);
	}
	free (ptr);
	ptr = empty_line;
    }
    return ptr;
}

char *get_filelist_line_short (void *data, int line_number, char buffer[1024]);

void cursor_moved (NeXTFileBrowser * browser, NeXTDir * dir, CWidget * list_w)
{
    if (browser && dir && list_w) {
	dir_save_position (dir, list_w->cursor, list_w->current, list_w->firstline);
	if (dir->cursor >= 0) {
	    struct file_entry *item = &(dir->list[dir->cursor]);
	    char buffer[512], *tail;
	    if (get_filelist_line_short (dir->list, dir->cursor, buffer))
		CRedrawDynText (browser->fileattr, buffer);
	}
    }
}

/*
   Returns "" on no file entered and NULL on exit (i.e. Cancel button pushed)
   else returns the file or directory. Result must be immediately copied.
   Result must not be free'd.
 */
char *handle_browser (const char *identifier, CEvent * cwevent, int options)
{
    int i;
    CWidget *w = CIdent (identifier);
    NeXTFileBrowser *browser;
    int flist = -1;
    char *ptr;

    if (!w)
	return empty_line;

    if (cwevent->command == CK_Cancel)
	return 0;
    if ((browser = (NeXTFileBrowser *) w->hook) == NULL)
	return empty_line;

    /* that will be our subwidget name : */
    if ((ptr = strrchr (cwevent->ident, '.')) == NULL)
	/* nothing to do here - messege for self */
	return empty_line;

    ptr++;
    /* lets find out which one we have (starting from the shortest) */
    if (strcmp (ptr, "ok") == 0)
	/* we take text from text input and use it as filename */
	return item_selected (identifier, browser, NULL);
    else if (strncmp (ptr, "flist", 5) == 0)
	/* all filelists ends with flistNNN */
	flist = atoi (ptr + 5);	/* will process them later */
    else if (strcmp (ptr, "cancel") == 0)
	return 0;
    else if (strcmp (ptr, "finp") == 0) {
	if (cwevent->type == KeyPress && cwevent->command == CK_Enter)
	    return item_selected (identifier, browser, NULL);
	return empty_line;
    } else if (strcmp (ptr, "filtinp") == 0) {
	if (cwevent->type == KeyPress && cwevent->command == CK_Enter)
	    filter_changed (browser, browser->tree->selected);
	return empty_line;
    }
    if (flist >= 0 && flist < browser->numpanes) {	/* filelist event */
	NeXTDir *dir = get_dir_by_num (browser->tree, flist + browser->tree->pos);
	CWidget *list_w = CIdent (browser->lists[flist]);

	/* saving status */
	browser->focused = flist;
	browser->tree->selected = dir;

	switch (cwevent->type) {
	case EnterNotify:
	    CFocus (list_w);
	    update_filter (browser, dir);
	    break;
/*          case LeaveNotify : 
   fprintf(stderr, "LeaveNotify\n");
   browser->focused = -1 ; 
   CFocus( CIdent(browser->name));
   break ;
 */
	case ButtonPress:
	case ButtonRelease:
	case MotionNotify:
	    cursor_moved (browser, dir, list_w);
	    if (cwevent->type != ButtonRelease)
		update_input (browser, dir);
	    else if (cwevent->double_click)
		return item_selected (identifier, browser, dir);
	    break;
	case KeyPress:
	    {
		int scroll = 0;
		switch (cwevent->command) {
		case CK_Enter:
		    update_input (browser, dir);
		    return item_selected (identifier, browser, dir);
		case CK_Up:
		case CK_Down:
		case CK_Page_Up:
		case CK_Page_Down:
		    cursor_moved (browser, dir, list_w);
		    break;
		case CK_Left:
		    if (browser->tree->pos > 0)
			scroll = -1;
		    break;
		case CK_Right:
		    if (browser->tree->pos < browser->tree->numdirs - browser->numpanes)
			scroll = 1;
		    break;
		case CK_Home:
		    scroll = -(browser->tree->pos);
		    break;
		case CK_End:
		    scroll = (browser->tree->numdirs - browser->numpanes) - browser->tree->pos;
		    break;
		case CK_BackSpace:
		    /* remove last typed partial filename char */
		    goto_partial_file_name (browser, dir, 0);
		    break;
		default:
		    if (cwevent->insert > 0) {
			if (cwevent->key == ' ')
			    update_input (browser, dir);
			else if (isascii (cwevent->key)) {
			    goto_partial_file_name (browser, dir, cwevent->key);
			    cursor_moved (browser, dir, list_w);
			}
		    }
		}
		if (scroll != 0) {
		    browser->tree->pos += scroll;
		    link_browser_to_data (browser, 0);
		    dir = get_dir_by_num (browser->tree, browser->tree->pos + flist);
		    update_filter (browser, dir);
		    update_scrollbar (w->hori_scrollbar, browser, 1);
		}
	    }
	    break;
	default:
	    break;
	}
    }
    return empty_line;
}

Window find_mapped_window (Window w);

/* result must be free'd */
char *get_file_or_dir (Window parent, int x, int y,
       const char *dir, const char *file, const char *label, int options)
{
    CEvent cwevent;
    XEvent xevent;
    CState s;
    CWidget *w;

    CBackupState (&s);
    CDisable ("*");
    CEnable ("_cfileBr*");

    parent = find_mapped_window (parent);
    if (!(x | y)) {
	x = 20;
	y = 20;
    }
    draw_file_browser ("CGetFile", parent, x, y, dir, file, label);

    CFocus (CIdent ("CGetFile.finp"));

    file = "";
    do {
	CNextEvent (&xevent, &cwevent);
	if (xevent.type == Expose || !xevent.type
	    || xevent.type == InternalExpose || xevent.type == TickEvent)
	    continue;
	if (!CIdent ("CGetFile")) {
	    file = 0;
	    break;
	}
	if (xevent.type == Expose || !xevent.type || xevent.type == AlarmEvent
	  || xevent.type == InternalExpose || xevent.type == TickEvent) {
	    file = "";
	    continue;
	}
	file = handle_browser ("CGetFile", &cwevent, options);
	if (!file)
	    break;
    } while (!(*file));

/* here we want to add the complete path to the text-input history: */
    w = CIdent ("CGetFile.finp");
    if (w) {
	if (w->text) {
	    free (w->text);
	    w->text = 0;
	}
	if (file)
	    w->text = (char *) strdup (file);
    }
    CDestroyWidget ("CGetFile");	/* text is added to history 
					   when text-input widget is destroyed */
    CRestoreState (&s);

    if (file)
	return (char *) ((*file) ? file : 0);
    else
	return 0;
}

/* 
   cb_browser():
   1) if filelist has decided that browser should be notifyed about event
   it sets CEvent.ident member, and we get into handle_browser.
   That includes events external for filelist (Left, Right, Home, 
   End, Space, Enter, Esc) even thou filelist does no processing of it.
   Note also that if user types in some characters while in filelist - 
   filelist will do search for filename, not the browser, as it was 
   happening in original version.
   2) handle_browser do all the neccessary processing of received events -
   updates edit boxes, updates NexTDirTree status, changes dir if needed,
   scrolls left, right, home and end.
   Sasha.
 */

char *get_browser_name (char *ident, char *buffer)
{
    char *start = buffer;
    while (*ident && *ident != '.')
	*(buffer++) = *(ident++);
    *buffer = '\0';

    return start;
}

int cb_browser (CWidget * w, XEvent * x, CEvent * c)
{
    char id[32];
    get_browser_name (w->ident, id);
    if (!handle_browser (id, c, GETFILE_BROWSER)) {
	w = CIdent (catstrs (id, ".finp", 0));
	if (w)
	    if (w->text) {
		free (w->text);
		w->text = 0;
	    }
	CDestroyWidget (id);
    }
    return 0;
}

void CDrawBrowser (const char *ident, Window parent, int x, int y,
		   const char *dir, const char *file, const char *label)
{
    CWidget *w;

    if (!(parent | x | y)) {
	parent = CFirstWindow;
	x = 20;
	y = 20;
    }
    draw_file_browser (ident, parent, x, y, dir, file, label);
    if ((w = CIdent (ident))) {
	NeXTFileBrowser *browser = w->hook;
	int i;
	if (browser) {
	    for (i = 0; i < browser->numpanes; i++)
		CAddCallback (browser->lists[i], cb_browser);

	    CAddCallback (browser->name, cb_browser);
	    CAddCallback (browser->filter, cb_browser);
	    CAddCallback (catstrs (ident, ".ok", 0), cb_browser);
	    CAddCallback (catstrs (ident, ".cancel", 0), cb_browser);

	    CFocus (CIdent (catstrs (ident, ".finp", 0)));
	}
    }
}

char *CGetFile (Window parent, int x, int y,
		const char *dir, const char *file, const char *label)
{
    return get_file_or_dir (parent, x, y, dir, file, label, 0);
}

char *CGetDirectory (Window parent, int x, int y,
		     const char *dir, const char *file, const char *label)
{
    return get_file_or_dir (parent, x, y, dir, file, label, GETFILE_GET_DIRECTORY);
}

char *CGetSaveFile (Window parent, int x, int y,
		    const char *dir, const char *file, const char *label)
{
    return get_file_or_dir (parent, x, y, dir, file, label, 0);
}

char *CGetLoadFile (Window parent, int x, int y,
		    const char *dir, const char *file, const char *label)
{
    return get_file_or_dir (parent, x, y, dir, file, label, GETFILE_GET_EXISTING_FILE);
}
