/* font.h - multiple font handling
   Copyright (C) 1999 Paul Sheer

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
 */

#include "coolwidget.h"
#include "coollocal.h"

#include "mad.h"

#define TEST_FONT_STRING "The Quick Brown Fox Jumps Over The Lazy Dog"

static struct font_stack {
    struct font_object *f;
    struct font_stack *next;
} *font_stack = 0;

int CTextWidth (const char *s, int l)
{
    if (!current_font->font_struct && current_font->font_set)
	return XmbTextEscapement (current_font->font_set, s, l);
    return XTextWidth (current_font->font_struct, s, l);
}

int CStringWidth (const char *s)
{
    return CTextWidth (s, strlen (s));
}

int CImageText (Window w, int x, int y, const char *s, int l)
{
    if (!current_font->font_struct && current_font->font_set) {
	XmbDrawImageString (CDisplay, w, current_font->font_set, CGC, x, y, s, l);
	return XmbTextEscapement (current_font->font_set, s, l);
    }
    return XDrawImageString (CDisplay, w, CGC, x, y, s, l);
}

#if 0
int CText (Window w, int x, int y, const char *s, int l)
{
    if (!current_font->font_struct && current_font->font_set) {
	XmbDrawString (CDisplay, w, current_font->font_set, CGC, x, y, s, l);
	return XmbTextEscapement (current_font->font_set, s, l);
    }
    return XDrawString (CDisplay, w, CGC, x, y, s, l);
}
#endif

int CImageString (Window w, int x, int y, const char *s)
{
    return CImageText (w, x, y, s, strlen (s));
}

static struct font_object *find_font (char *name)
{
    struct font_object *n;
    if (current_font)
	if (!strcmp (current_font->name, name))
	    return current_font;
    for (n = all_fonts; n; n = n->next)
	if (!strcmp (n->name, name))
	    return n;
    return 0;
}

/* returns width. the descent is only ever used for drawing an underbar.
   the ascent is only ever used in calculating FONT_PIX_PER_LINE */
static int get_string_dimensions (char *s, int n, int *height, int *ascent, int *ink_descent)
{
    int width, direction;
    XRectangle ink;
    XRectangle logical;
    if (!current_font->font_struct && current_font->font_set) {
	width = XmbTextExtents (current_font->font_set, s, n, &ink, &logical);
	if (height)
	    *height = logical.height;
	if (ascent)
	    *ascent = (-logical.y);
	if (ink_descent)
	    *ink_descent = ink.height + ink.y;
    } else {
	XCharStruct c;
	int ascent_, descent;
	XTextExtents (current_font->font_struct, s, n, &direction, &ascent_, &descent, &c);
	if (ascent)
	    *ascent = ascent_;
	if (height)
	    *height = ascent_ + descent;
	if (ink_descent) {
	    if (n == 1) {
		int i;
		i = (unsigned char) *s;
		if (i >= current_font->font_struct->min_char_or_byte2
		    && i <= current_font->font_struct->max_char_or_byte2)
			*ink_descent =
			current_font->font_struct->per_char[i -
							    current_font->font_struct->
							    min_char_or_byte2].descent;
		else
		    *ink_descent = 0;
	    } else
		*ink_descent = descent;
	}
	width = CTextWidth (s, n);
    }
    return width;
}

static int check_font_fixed (void)
{
    int m;
    char *p;
    m = get_string_dimensions ("M", 1, 0, 0, 0);
    for (p = "!MI .1@~"; *p; p++)
	if (m != get_string_dimensions (p, 1, 0, 0, 0))
	    return 0;
    return m;
}

static void get_font_dimensions (void)
{
    unsigned char *p, q[256];
    int i;

/* fill an array with graphical characters of the current locale */
    for (p = q, i = 1; i < 255; i++)
	if (isgraph ((unsigned char) i))
	    *p++ = (unsigned char) i;
    *p = '\0';

/* get the dimensions of the complete character set */
    get_string_dimensions ((char *) q, i, &FONT_HEIGHT, &FONT_ASCENT, &FONT_DESCENT);

/* create an example sentance */
    p = (unsigned char *) TEST_FONT_STRING;

/* get the mean font width from the example sentance */
    FONT_MEAN_WIDTH = get_string_dimensions ((char *) p, strlen ((char *) p), 0, 0, 0) / strlen ((char *) p);

/* now for each character calculate its descent and width */
    for (i = 0; i < 256; i++) {
	char t;
	int d;
	t = (unsigned char) i;
	FONT_PER_CHAR[i] = get_string_dimensions (&t, 1, 0, 0, &d);
	FONT_PER_CHAR_DESCENT[i] = (unsigned char) d;
	if (FIXED_FONT && (FONT_PER_CHAR[i] != FIXED_FONT)) {
	    FONT_PER_CHAR[i] = 0;	/* zero bad characters of fix fonts */
	    FONT_PER_CHAR_DESCENT[i] = 0;
	}
    }
}

/* seems you cannot draw different fonts in the same GC. this is
   strange???, so we create a dummy GC for each font */
static Window get_dummy_gc (void)
{
    static Window dummy_window = 0;
    XGCValues gcv;
    if (!dummy_window) {
	XSetWindowAttributes xswa;
	xswa.override_redirect = 1;
	dummy_window = XCreateWindow (CDisplay, CRoot, 0, 0, 1, 1, 0, CDepth, InputOutput, CVisual, CWOverrideRedirect, &xswa);
    }
    gcv.foreground = COLOR_BLACK;
    if (current_font->font_struct) {
	gcv.font = current_font->font_struct->fid;
	CGC = XCreateGC (CDisplay, dummy_window, GCForeground | GCFont, &gcv);
    } else {
	CGC = XCreateGC (CDisplay, dummy_window, GCForeground, &gcv);
    }
    return dummy_window;
}

static XFontSet get_font_set (char *name)
{
    XFontSet fontset;
    char **a = 0;
    int b;
    if (!XSupportsLocale ())
	fprintf (stderr, "X does not support the locale: %s\n", setlocale (LC_CTYPE, 0));
    fontset = XCreateFontSet (CDisplay, name, &a, &b, 0);
    if (!fontset)
	return 0;
    XFreeStringList (a);
    return fontset;
}

/* loads a font and sets the current font to it */
static struct font_object *load_font (char *name, char *xname)
{
    struct font_object *n;
    Window w;
    n = CMalloc (sizeof (struct font_object));
    memset (n, 0, sizeof (struct font_object));
    n->font_set = get_font_set (xname);
    if (!n->font_set)
	fprintf (stderr, _ ("%s: display %s cannot load font\n\t%s\nas a font set - trying raw load.\n"),
		 CAppName, DisplayString (CDisplay), xname);
/* font set may have failed only because of an invalid locale, but
   we try load the font anyway */
    if (!n->font_set && !strchr (xname, ',')) {
	n->font_struct = XLoadQueryFont (CDisplay, xname);
	n->free_font_struct = 1;
    }
/* font set may have worked, but if there is only one font, we
   might as well try use a font struct (which draws faster) */
    if (n->font_set && !n->font_struct) {
	int i, num_fonts, single_font = 1;
	XFontStruct **font_struct_list_return = 0;
	char **font_name_list_return = 0;
	num_fonts = XFontsOfFontSet (n->font_set, &font_struct_list_return, &font_name_list_return);
/* check if there is really only one font */
	for (i = 1; i < num_fonts; i++) {
	    if (strcmp (font_name_list_return[0], font_name_list_return[1])) {
		single_font = 0;
		break;
	    }
	}
	if (single_font) {
	    n->font_struct = font_struct_list_return[0];
	    n->free_font_struct = 0;
	}
    }
    if (!n->font_struct && !n->font_set) {
	fprintf (stderr, _ ("%s: display %s cannot load font\n\t%s\n"),
		 CAppName, DisplayString (CDisplay), xname);
	free (n);
	return 0;
    }
    n->next = all_fonts;
    current_font = all_fonts = n;
    n->name = (char *) strdup (name);
    w = get_dummy_gc ();
    if (current_font->font_set)
	XmbDrawImageString (CDisplay, w, current_font->font_set, CGC, 0, 0, " AZ~", 4);
    if (current_font->font_struct)
	XDrawImageString (CDisplay, w, CGC, 0, 0, " AZ~", 4);
    FIXED_FONT = check_font_fixed ();
    get_font_dimensions ();
/* -----> FIXME: yes, you actually HAVE to draw with the GC for
   the font to stick. can someone tell me what I am doing wrong? */
    return n;
}

int CPushFont (char *name, char *xname)
{
    struct font_stack *p;
    struct font_object *f;
    f = find_font (name);
    if (f) {
	f->ref++;
    } else {
	f = load_font (name, xname);
	if (!f)
	    return 1;
	f->ref = 1;
    }
    p = CMalloc (sizeof (struct font_stack));
    p->f = f;
    p->next = font_stack;
    font_stack = p;
    current_font = font_stack->f;
    return 0;
}

void CPopFont (void)
{
    struct font_stack *p;
    if (!font_stack) {
	fprintf (stderr, "Huh\n?");
	abort ();
    }
    if (!--font_stack->f->ref) {
	XFreeGC (CDisplay, font_stack->f->gc);
	if (font_stack->f->font_set)
	    XFreeFontSet (CDisplay, font_stack->f->font_set);
	if (font_stack->f->font_struct)
	    if (font_stack->f->free_font_struct)
		XFreeFont (CDisplay, font_stack->f->font_struct);
	free (font_stack->f->name);
	free (font_stack->f);
    }
    p = font_stack->next;
    if (p) {
	current_font = p->f;
    } else {
	current_font = 0;
    }
    free (font_stack);
    font_stack = p;
}

void CFreeAllFonts (void)
{
    int i = 0;
    while (font_stack) {
	CPopFont ();
	i++;
    }
}

int CIsFixedFont (void)
{
    return FIXED_FONT;
}

